@@ -8,28 +8,28 @@ Functions and general utilities for the deepof package.
 
 """
 
-import argparse
-import multiprocessing
-import os
 from copy import deepcopy
-from itertools import combinations, product
-from typing import Tuple, Any, List, Union, NewType
-
-import cv2
-import math
-import networkx as nx
-import numpy as np
-import pandas as pd
-import regex as re
-import ruptures as rpt
-import tensorflow as tf
 from dask_image.imread import imread
+from itertools import combinations, product
 from joblib import Parallel, delayed
 from scipy.signal import savgol_filter
 from shapely.geometry import Polygon
 from sklearn import mixture
 from sklearn.feature_selection import VarianceThreshold
+from sklearn.preprocessing import StandardScaler, MinMaxScaler, RobustScaler
 from tqdm import tqdm
+from typing import Tuple, Any, List, Union, NewType
+import argparse
+import cv2
+import math
+import multiprocessing
+import networkx as nx
+import numpy as np
+import os
+import pandas as pd
+import regex as re
+import ruptures as rpt
+import tensorflow as tf
 
 # DEFINE CUSTOM ANNOTATED TYPES #
 project = NewType("deepof_project", Any)
@@ -37,44 +37,117 @@ coordinates = NewType("deepof_coordinates", Any)
 table_dict = NewType("deepof_table_dict", Any)
 
 
-# CONNECTIVITY FOR DLC MODELS
+# CONNECTIVITY AND GRAPH REPRESENTATIONS
 
 
-def connect_mouse_topview(animal_id=None) -> nx.Graph:
+def connect_mouse_topview(animal_ids=None, exclude_bodyparts: list = None) -> nx.Graph:
     """
 
     Creates a nx.Graph object with the connectivity of the bodyparts in the
     DLC topview model for a single mouse. Used later for angle computing, among others.
 
     Args:
-        animal_id (str): if more than one animal is tagged, specify the animal identyfier as a string.
+        animal_ids (str): if more than one animal is tagged, specify the animal identyfier as a string.
+        exclude_bodyparts (list): Remove the specified nodes from the graph.
 
     Returns:
         connectivity (nx.Graph)
 
     """
 
-    connectivity = {
-        "Nose": ["Left_ear", "Right_ear", "Spine_1"],
-        "Left_ear": ["Right_ear", "Spine_1"],
-        "Right_ear": ["Spine_1"],
-        "Spine_1": ["Center", "Left_fhip", "Right_fhip"],
-        "Center": ["Left_fhip", "Right_fhip", "Spine_2", "Left_bhip", "Right_bhip"],
-        "Spine_2": ["Left_bhip", "Right_bhip", "Tail_base"],
-        "Tail_base": ["Tail_1", "Left_bhip", "Right_bhip"],
-        "Tail_1": ["Tail_2"],
-        "Tail_2": ["Tail_tip"],
-    }
+    if animal_ids is None:
+        animal_ids = [""]
+    if not isinstance(animal_ids, list):
+        animal_ids = [animal_ids]
 
-    connectivity = nx.Graph(connectivity)
+    connectivities = []
 
-    if animal_id:
-        mapping = {
-            node: "{}_{}".format(animal_id, node) for node in connectivity.nodes()
+    for animal_id in animal_ids:
+        connectivity = {
+            "Nose": ["Left_ear", "Right_ear"],
+            "Spine_1": ["Center", "Left_ear", "Right_ear"],
+            "Center": ["Left_fhip", "Right_fhip", "Spine_2"],
+            "Spine_2": ["Left_bhip", "Right_bhip", "Tail_base"],
+            "Tail_base": ["Tail_1"],
+            "Tail_1": ["Tail_2"],
+            "Tail_2": ["Tail_tip"],
         }
-        nx.relabel_nodes(connectivity, mapping, copy=False)
+        connectivity = nx.Graph(connectivity)
 
-    return connectivity
+        if animal_id:
+            mapping = {
+                node: "{}_{}".format(animal_id, node) for node in connectivity.nodes()
+            }
+            if exclude_bodyparts is not None:
+                exclude = ["{}_{}".format(animal_id, exc) for exc in exclude_bodyparts]
+            nx.relabel_nodes(connectivity, mapping, copy=False)
+        else:
+            exclude = exclude_bodyparts
+
+        if exclude_bodyparts is not None:
+            connectivity.remove_nodes_from(exclude)
+
+        connectivities.append(connectivity)
+
+    if len(connectivities) > 1:
+        pass
+
+    final_graph = connectivities[0]
+    for g in range(1, len(connectivities)):
+        final_graph = nx.compose(final_graph, connectivities[g])
+        final_graph.add_edge(
+            "{}_Nose".format(animal_ids[g - 1]), "{}_Nose".format(animal_ids[g])
+        )
+        final_graph.add_edge(
+            "{}_Tail_base".format(animal_ids[g - 1]),
+            "{}_Tail_base".format(animal_ids[g]),
+        )
+        final_graph.add_edge(
+            "{}_Nose".format(animal_ids[g]), "{}_Tail_base".format(animal_ids[g - 1])
+        )
+        final_graph.add_edge(
+            "{}_Nose".format(animal_ids[g - 1]), "{}_Tail_base".format(animal_ids[g])
+        )
+
+    return final_graph
+
+
+def edges_to_weithed_adj(adj: np.ndarray, edges: np.ndarray):
+    """Converts an edge feature matrix to a weighted adjacency matrix.
+
+    Args:
+        - adj (np.ndarray): binary adjacency matrix of the current graph.
+        - edges (np.ndarray): edge feature matrix. Last two axes should be of shape nodes x features.
+
+    """
+    adj = np.repeat(np.expand_dims(adj.astype(float), axis=0), edges.shape[0], axis=0)
+    if len(edges.shape) == 3:
+        adj = np.repeat(np.expand_dims(adj, axis=1), edges.shape[1], axis=1)
+
+    adj[np.where(adj)] = np.concatenate([edges, edges[:, ::-1]], axis=-2).flatten()
+
+    return adj
+
+
+def enumerate_all_bridges(G: nx.graph) -> list:
+    """Enumerates all 3-node connected sequences in the given graph.
+
+    Args:
+        - G (nx.graph): Animal connectivity graph.
+
+    Returns:
+        bridges (list): List with all 3-node connected sequences in the provided graph.
+
+    """
+    degrees = dict(nx.degree(G))
+    centers = [node for node in degrees.keys() if degrees[node] >= 2]
+
+    bridges = []
+    for center in centers:
+        for comb in list(combinations(list(G[center].keys()), 2)):
+            bridges.append([comb[0], center, comb[1]])
+
+    return bridges
 
 
 # QUALITY CONTROL AND PREPROCESSING #
@@ -222,20 +295,19 @@ def bpart_distance(
     return pd.concat(dists, axis=1)
 
 
-def angle(a: np.array, b: np.array, c: np.array) -> np.array:
+def angle(bpart_array: np.array) -> np.array:
     """
 
-    Returns a numpy.array with the angles between the provided instances.
+    Returns a numpy.ndarray with the angles between the provided instances.
 
     Args:
-        a (np.array): 2D positions over time for a body part.
-        b (np.array): 2D positions over time for a body part.
-        c (np.array): 2D positions over time for a body part.
+        bpart_array (numpy.array): 2D positions over time for a bodypart.
 
     Returns:
         ang (np.array): 1D angles between the three-point-instances.
 
     """
+    a, b, c = bpart_array
 
     ba = a - b
     bc = c - b
@@ -248,25 +320,6 @@ def angle(a: np.array, b: np.array, c: np.array) -> np.array:
     return ang
 
 
-def angle_trio(bpart_array: np.array) -> np.array:
-    """
-
-    Returns a numpy.array with all three possible angles between the provided instances.
-
-    Args:
-        bpart_array (numpy.array): 2D positions over time for a bodypart.
-
-    Returns:
-        ang_trio (numpy.array): 2D all-three angles between the three-point-instances.
-
-    """
-
-    a, b, c = bpart_array
-    ang_trio = np.array([angle(a, b, c), angle(a, c, b), angle(b, a, c)])
-
-    return ang_trio
-
-
 def compute_areas(coords, animal_id=None):
     """
     Computes relevant areas (head, torso, back, full) for the provided coordinates.
@@ -306,7 +359,10 @@ def compute_areas(coords, animal_id=None):
         x = coords.xs(key="x", level=1)[bps]
         y = coords.xs(key="y", level=1)[bps]
 
-        areas.append(Polygon(zip(x, y)).area)
+        if np.isnan(x).any() or np.isnan(y).any():
+            areas.append(np.nan)
+        else:
+            areas.append(Polygon(zip(x, y)).area)
 
     return areas
 
@@ -383,6 +439,81 @@ def align_trajectories(data: np.array, mode: str = "all") -> np.array:
     return aligned_trajs
 
 
+def scale_table(
+    coordinates: coordinates,
+    feature_array: np.ndarray,
+    scale: str,
+    global_scaler: Any = None,
+):
+    """Scales features in a table controlling for both individual body size and interanimal variability.
+
+    Args:
+        coordinates (coordinates): a deepof coordinates object.
+        feature_array (np.ndarray): array to scale. Should be shape (instances x features).
+        scale (str): Data scaling method. Must be one of 'standard', 'robust' (default; recommended) and 'minmax'.
+        global_scaler (Any): global scaler, fit in the whole dataset.
+
+    """
+    exp_temp = feature_array.to_numpy()
+
+    annot_length = 0
+    if coordinates._propagate_labels:
+        exp_temp = exp_temp[:, :-1]
+        annot_length += 1
+
+    if coordinates._propagate_annotations:
+        exp_temp = exp_temp[
+            :, : -list(coordinates._propagate_annotations.values())[0].shape[1]
+        ]
+        annot_length += list(coordinates._propagate_annotations.values())[0].shape[1]
+
+    if global_scaler is None:
+        # Scale each modality separately using a custom function
+        exp_temp = scale_animal(exp_temp, scale)
+    else:
+        # Scale all experiments together, to control for differential stats
+        exp_temp = global_scaler.transform(exp_temp)
+
+    current_tab = np.concatenate(
+        [
+            exp_temp,
+            feature_array.copy().to_numpy()[:, feature_array.shape[1] - annot_length :],
+        ],
+        axis=1,
+    )
+
+    return current_tab
+
+
+def scale_animal(feature_array: np.ndarray, scale: str):
+    """Scales features in the provided array.
+
+    Args:
+        feature_array (np.ndarray): array to scale. Should be shape (instances x features).
+        graph (nx.Graph): connectivity graph for the current animals.
+        scale (str): Data scaling method. Must be one of 'standard', 'robust' (default; recommended) and 'minmax'.
+
+    Returns:
+        Scaled version of the input array, with features normalized by modality.
+        List of scalers per modality.
+
+    """
+    scalers = []
+
+    # number of body part sets to use for coords (x, y), speeds, and distances
+    if scale == "standard":
+        cur_scaler = StandardScaler()
+    elif scale == "minmax":
+        cur_scaler = MinMaxScaler()
+    else:
+        cur_scaler = RobustScaler()
+
+    normalized_array = cur_scaler.fit_transform(feature_array)
+    scalers.append(cur_scaler)
+
+    return normalized_array
+
+
 def kleinberg(offsets, s=np.e, gamma=1.0, n=None, T=None, k=None):
     """Kleinberg's algorithm (described in 'Bursty and Hierarchical Structure
     in Streams'). The algorithm models activity bursts in a time series as an
@@ -441,7 +572,7 @@ def kleinberg(offsets, s=np.e, gamma=1.0, n=None, T=None, k=None):
         else:
             return (j - i) * gamma_log_n
 
-    alpha_function = np.vectorize(lambda x: s ** x / g_hat)
+    alpha_function = np.vectorize(lambda x: s**x / g_hat)
     alpha = alpha_function(np.arange(k))
 
     def f(j, x):
@@ -674,11 +805,7 @@ def rupture_per_experiment(
         if i in rupture_indices:
             current_size = tab.shape[0]
             current_train, current_breaks = rolling_window(
-                (
-                    to_rupture[cumulative_shape : cumulative_shape + current_size]
-                    if automatic_changepoints
-                    else to_rupture
-                ),
+                to_rupture[cumulative_shape : cumulative_shape + current_size],
                 window_size,
                 window_step,
                 automatic_changepoints,
@@ -688,7 +815,8 @@ def rupture_per_experiment(
             # to avoid skipping breakpoints between experiments
             if current_breaks is not None:
                 current_breaks = np.array(current_breaks) + cumulative_shape
-                cumulative_shape += current_size
+
+            cumulative_shape += current_size
 
             try:  # pragma: no cover
                 # To concatenate the current ruptures with the ones obtained
@@ -714,8 +842,7 @@ def rupture_per_experiment(
                     )
 
                 # Once that's taken care of, concatenate ruptures alongside axis 0
-                if automatic_changepoints:
-                    ruptured_dataset = np.concatenate([ruptured_dataset, current_train])
+                ruptured_dataset = np.concatenate([ruptured_dataset, current_train])
                 if current_breaks is not None:
                     break_indices = np.concatenate([break_indices, current_breaks])
             except (ValueError, AttributeError):
@@ -841,7 +968,7 @@ def full_outlier_mask(
     replaced by a boolean mask, where True indicates an outlier.
 
     Args:
-        experiment (pd.DataFrame): Data frame with time series representing the x, y positions of a every body part
+        experiment (pd.DataFrame): Data frame with time series representing the x, y positions of every body part
         likelihood (pd.DataFrame): Data frame with likelihood data per body part as extracted from deeplabcut
         likelihood_tolerance (float): Minimum tolerated likelihood, below which an outlier is called
         exclude (str): Body part to exclude from the analysis (to concatenate with bpart alignment)
@@ -894,7 +1021,7 @@ def interpolate_outliers(
     Note that this approach only works for equally spaced data (constant camera acquisition rates).
 
     Args:
-        experiment (pd.DataFrame): Data frame with time series representing the x, y positions of a every body part.
+        experiment (pd.DataFrame): Data frame with time series representing the x, y positions of every body part.
         likelihood (pd.DataFrame): Data frame with likelihood data per body part as extracted from deeplabcut.
         likelihood_tolerance (float): Minimum tolerated likelihood, below which an outlier is called.
         exclude (str): Body part to exclude from the analysis (to concatenate with bpart alignment).
@@ -910,6 +1037,7 @@ def interpolate_outliers(
 
     interpolated_exp = experiment.copy()
 
+    # Creates a mask marking all outliers
     mask = full_outlier_mask(
         experiment, likelihood, likelihood_tolerance, exclude, lag, n_std, mode
     )
@@ -967,14 +1095,11 @@ def automatically_recognize_arena(
     path: str = ".",
     tables: dict = None,
     recoglimit: int = 500,
-    high_fidelity: bool = False,
     arena_type: str = "circular-autodetect",
-    detection_mode: str = "rule-based",
-    cnn_model: tf.keras.models.Model = None,
 ) -> Tuple[np.array, int, int]:
     """
 
-    Returns numpy.array with information about the arena recognised from the first frames
+    Returns numpy.ndarray with information about the arena recognised from the first frames
     of the video. WARNING: estimates won't be reliable if the camera moves along the video.
 
     Args:
@@ -983,13 +1108,8 @@ def automatically_recognize_arena(
         path (str): Full path of the directory where the videos are.
         tables (dict): Dictionary with DLC time series in DataFrames as values.
         recoglimit (int): Number of frames to use for position estimates.
-        high_fidelity (bool): If True, runs arena recognition on the whole video. Slow, but
         potentially more accurate in poor lighting conditions.
         arena_type (string): Arena type; must be one of ['circular-autodetect', 'circular-manual', 'polygon-manual'].
-        detection_mode (str): Algorithm to use to detect the arena. "cnn" uses a
-        pretrained model based on ResNet50 to predict the ellipse parameters from
-        the image. "rule-based" (default) uses a simpler (and faster) image segmentation approach.
-        cnn_model (tf.keras.models.Model): Model to use if detection_mode=="cnn".
 
     Returns:
         arena (np.ndarray): 1D-array containing information about the arena.
@@ -1003,9 +1123,6 @@ def automatically_recognize_arena(
 
     cap = cv2.VideoCapture(os.path.join(path, videos[vid_index]))
 
-    if high_fidelity:
-        recoglimit = int(1e10)  # set recoglimit to a very big value
-
     if tables is not None:
         # Select relevant table to check animal positions; if animals are close to the arena, do not take those frames
         # into account
@@ -1033,9 +1150,7 @@ def automatically_recognize_arena(
         if arena_type == "circular-autodetect":
 
             # Detect arena and extract positions
-            temp_center, temp_axes, temp_angle = circular_arena_recognition(
-                frame, detection_mode=detection_mode, cnn_model=cnn_model
-            )
+            temp_center, temp_axes, temp_angle = circular_arena_recognition(frame)
             temp_arena = np.array([[*temp_center, *temp_axes, temp_angle]])
 
             # Set if not assigned, else concat and return the median
@@ -1074,7 +1189,9 @@ def automatically_recognize_arena(
     return arena, h, w
 
 
-def retrieve_corners_from_image(frame: np.ndarray, arena_type: str):  # pragma: no cover
+def retrieve_corners_from_image(
+    frame: np.ndarray, arena_type: str, cur_vid: int, videos: list
+):  # pragma: no cover
     """
 
     Opens a window and waits for the user to click on all corners of the polygonal arena.
@@ -1083,6 +1200,8 @@ def retrieve_corners_from_image(frame: np.ndarray, arena_type: str):  # pragma:
     Args:
         frame (np.ndarray): Frame to display.
         arena_type (str): Type of arena to be used. Must be one of the following: "circular-manual", "polygon-manual".
+        cur_vid (int): Index of the current video in the list of videos.
+        videos (list): List of videos to be processed.
 
     Returns:
 
@@ -1106,11 +1225,15 @@ def retrieve_corners_from_image(frame: np.ndarray, arena_type: str):  # pragma:
         frame_copy = frame.copy()
 
         cv2.imshow(
-            "deepof - Select polygonal arena corners - (q: exit / d: delete)",
+            "deepof - Select polygonal arena corners - (q: exit / d: delete) - {}/{} processed".format(
+                cur_vid, len(videos)
+            ),
             frame_copy,
         )
         cv2.setMouseCallback(
-            "deepof - Select polygonal arena corners - (q: exit / d: delete)",
+            "deepof - Select polygonal arena corners - (q: exit / d: delete) - {}/{} processed".format(
+                cur_vid, len(videos)
+            ),
             click_on_corners,
         )
 
@@ -1150,7 +1273,9 @@ def retrieve_corners_from_image(frame: np.ndarray, arena_type: str):  # pragma:
             )
 
         cv2.imshow(
-            "deepof - Select polygonal arena corners - (q: exit / d: delete)",
+            "deepof - Select polygonal arena corners - (q: exit / d: delete) - {}/{} processed".format(
+                cur_vid, len(videos)
+            ),
             frame_copy,
         )
 
@@ -1158,11 +1283,12 @@ def retrieve_corners_from_image(frame: np.ndarray, arena_type: str):  # pragma:
         if cv2.waitKey(1) & 0xFF == ord("d"):
             corners = corners[:-1]
 
-        # Exit is user presses q
-        elif cv2.waitKey(1) & 0xFF == ord("q"):
-            cv2.destroyAllWindows()
-            cv2.waitKey(1)
-            break
+        # Exit is user presses 'q'
+        if len(corners) > 2:
+            if cv2.waitKey(1) & 0xFF == ord("q"):
+                for i in range(1, 5):
+                    cv2.waitKey(1)
+                break
 
     cv2.destroyAllWindows()
     cv2.waitKey(1)
@@ -1172,7 +1298,7 @@ def retrieve_corners_from_image(frame: np.ndarray, arena_type: str):  # pragma:
 
 
 def extract_polygonal_arena_coordinates(
-    video_path: str, arena_type: str
+    video_path: str, arena_type: str, video_index: int, videos: list
 ):  # pragma: no cover
     """
 
@@ -1182,6 +1308,8 @@ def extract_polygonal_arena_coordinates(
     Args:
         video_path: Path to the video file.
         arena_type: Type of arena to be used. Must be one of the following: "circular-manual", "polygon-manual".
+        video_index: Index of the current video in the list of videos.
+        videos: List of videos to be processed.
 
     Returns:
         np.ndarray: nx2 array containing the x-y coordinates of all n corners of the polygonal arena.
@@ -1195,7 +1323,10 @@ def extract_polygonal_arena_coordinates(
 
     # Get and return the corners of the arena
     arena_corners = retrieve_corners_from_image(
-        current_video[current_frame].compute(), arena_type
+        current_video[current_frame].compute(),
+        arena_type,
+        video_index,
+        videos,
     )
     return arena_corners, current_video.shape[2], current_video.shape[1]
 
@@ -1228,8 +1359,6 @@ def fit_ellipse_to_polygon(polygon: list):  # pragma: no cover
 
 def circular_arena_recognition(
     frame: np.ndarray,
-    detection_mode: str = "rule-based",
-    cnn_model: tf.keras.models.Model = None,
 ) -> np.array:
     """
 
@@ -1238,10 +1367,6 @@ def circular_arena_recognition(
 
     Args:
         frame (np.ndarray): numpy.ndarray representing an individual frame of a video
-        detection_mode (str): Algorithm to use to detect the arena. "cnn" uses a
-        pretrained model based on ResNet50 to predict the ellipse parameters from
-        the image. "rule-based" uses a simpler (and faster) image segmentation approach.
-        cnn_model (tf.keras.models.Model): Model to use if detection_mode=="cnn".
 
     Returns:
         circles (np.ndarray): 3-element-array containing x,y positions of the center
@@ -1249,47 +1374,22 @@ def circular_arena_recognition(
 
     """
 
-    if detection_mode == "rule-based":
+    # Convert image to grayscale, threshold it and close it with a 5x5 kernel
+    kernel = np.ones((5, 5))
+    gray_image = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
+    ret, thresh = cv2.threshold(gray_image, 255 // 4, 255, 0)
+    for _ in range(5):
+        thresh = cv2.morphologyEx(thresh, cv2.MORPH_CLOSE, kernel)
 
-        # Convert image to grayscale, threshold it and close it with a 5x5 kernel
-        kernel = np.ones((5, 5))
-        gray_image = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
-        ret, thresh = cv2.threshold(gray_image, 255 // 4, 255, 0)
-        for _ in range(5):
-            thresh = cv2.morphologyEx(thresh, cv2.MORPH_CLOSE, kernel)
+    # Obtain contours from the image, and retain the largest one
+    cnts, _ = cv2.findContours(
+        thresh.astype(np.int64), cv2.RETR_FLOODFILL, cv2.CHAIN_APPROX_TC89_KCOS
+    )
+    main_cnt = np.argmax([len(c) for c in cnts])
 
-        # Obtain contours from the image, and retain the largest one
-        cnts, _ = cv2.findContours(
-            thresh.astype(np.int64), cv2.RETR_FLOODFILL, cv2.CHAIN_APPROX_TC89_KCOS
-        )
-        main_cnt = np.argmax([len(c) for c in cnts])
-
-        center_coordinates, axes_length, ellipse_angle = fit_ellipse_to_polygon(
-            cnts[main_cnt]
-        )
-
-    elif detection_mode == "cnn":
-
-        input_shape = tuple(cnn_model.input.shape[1:-1])
-        image_temp = cv2.resize(frame, input_shape)
-        image_temp = image_temp / 255
-
-        # Detect the main ellipse containing the arena
-        predicted_arena = cnn_model.predict(image_temp[np.newaxis, :])[0]
-
-        # Parameters to return
-        center_coordinates = tuple(
-            (predicted_arena[:2] * frame.shape[:2][::-1] / input_shape).astype(int)
-        )
-        axes_length = tuple(
-            (predicted_arena[2:4] * frame.shape[:2][::-1] / input_shape).astype(int)
-        )
-        ellipse_angle = predicted_arena[4]
-
-    else:
-        raise ValueError(
-            "Invalid detection mode. Select between 'cnn' and 'rule-based'"
-        )
+    center_coordinates, axes_length, ellipse_angle = fit_ellipse_to_polygon(
+        cnts[main_cnt]
+    )
 
     # noinspection PyUnboundLocalVariable
     return center_coordinates, axes_length, ellipse_angle
@@ -1325,13 +1425,10 @@ def rolling_speed(
     """
 
     original_shape = dframe.shape
-    if center:
-        body_parts = [bp for bp in dframe.columns.levels[0] if center not in bp]
-    else:
-        try:
-            body_parts = dframe.columns.levels[0]
-        except AttributeError:
-            body_parts = dframe.columns
+    try:
+        body_parts = dframe.columns.levels[0]
+    except AttributeError:
+        body_parts = dframe.columns
 
     speeds = pd.DataFrame
 
@@ -1366,6 +1463,56 @@ def rolling_speed(
     return speeds.fillna(0.0)
 
 
+def filter_short_bouts(
+    cluster_assignments: np.ndarray,
+    cluster_confidence: np.ndarray,
+    confidence_indices: np.ndarray,
+    min_confidence: float = 0.0,
+    min_bout_duration: int = None,
+):  # pragma: no cover
+    """Filters out cluster assignment bouts shorter than min_bout_duration.
+
+    Args:
+        cluster_assignments (np.ndarray): Array of cluster assignments.
+        cluster_confidence (np.ndarray): Array of cluster confidence values.
+        confidence_indices (np.ndarray): Array of confidence indices.
+        min_confidence (float): Minimum confidence value.
+        min_bout_duration (int): Minimum bout duration in frames.
+
+    Returns:
+        np.ndarray: Mask of confidence indices to keep.
+
+    """
+    # Compute bout lengths, and filter out bouts shorter than min_bout_duration
+    bout_lengths = np.diff(
+        np.where(
+            np.diff(np.concatenate([[np.inf], cluster_assignments, [np.inf]])) != 0
+        )[0]
+    )
+
+    if min_bout_duration is None:
+        min_bout_duration = np.mean(bout_lengths)
+
+    confidence_indices[
+        np.repeat(bout_lengths, bout_lengths) < min_bout_duration
+    ] = False
+
+    # Compute average confidence per bout
+    cum_bout_lengths = np.concatenate([[0], np.cumsum(bout_lengths)])
+    bout_average_confidence = np.array(
+        [
+            cluster_confidence[confidence_indices][
+                cum_bout_lengths[i] : cum_bout_lengths[i + 1]
+            ].mean()
+            for i in range(len(bout_lengths))
+        ]
+    )
+
+    return (np.repeat(bout_average_confidence, bout_lengths) >= min_confidence) & (
+        confidence_indices
+    )
+
+
 # MACHINE LEARNING FUNCTIONS #
 
 
